function [L, X] = dpr1eig(D, u, v, r, tol)
% Function dpr1eig eigenpairs of the matrix
%   A = diag(D) + r * u * v',
% where D, u, v are n dimenisonal complex vectors, and r is a scalar
% If not needed eigenvectors are not calculated.
%
% INPUT:  
%   D, u, v:  vectors
%   r:        scalar r, 
%   tol:      structure of tolerance values
% Output: 
%   L:        eigenvalues
%   X:        eigenvectors
%
% =========================================================================
% |  dpr1eig.m                                                            |
% |  Copyright (C) 2020 Matea Ugrica, Tim Mitchell                        |
% |                                                                       |
% |  This program is free software: you can redistribute it and/or modify |
% |  it under the terms of the GNU Affero General Public License as       |
% |  published by the Free Software Foundation, either version 3 of       |
% |  the License, or (at your option) any later version.                  |
% |                                                                       |
% |  This program is distributed in the hope that it will be useful,      |
% |  but WITHOUT ANY WARRANTY; without even the implied warranty of       |
% |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        |
% |  GNU Affero General Public License for more details.                  |
% |                                                                       |
% |  You should have received a copy of the GNU Affero General Public     |
% |  License along with this program.  If not, see                        |
% |  <http://www.gnu.org/licenses/agpl.html>.                             |
% =========================================================================

    % Original algorithm solves eigenproblem for the matrix
    %   diag(D) + r * u * transpose(v).
    % So we need to conjugate v.
    if nargout == 2
        [L, X] = dpr1eig_callback(D, u, conj(v), r,tol);  
    else
        L = dpr1eig_callback(D, u, conj(v), r,tol); 
    end
end

function [LN, XN] = dpr1eig_callback(D, u, v, r,tol)
    flag_break = 0;
    
    LN = D;    
    % making symmetric dpr1 matrix
    s = sqrt(v ./ u);
    v = s .* u;
    
    % Need to get rid of nans in v but the corresponding eigenvalues for
    % these can just be read off the diagonal 
    v(isnan(v)) = 0;
    
    ind = find(abs(v.*v)/max(abs(D)) > tol.zero);
    
    s = s(ind);
    v = v(ind);
    D = D(ind);
    
    N = size(D, 1);     % Dimension
    L = zeros(N, 1);    % Eigenvalues
    
    BD = D; Bv = v; l = N;

    num_iter = zeros(N,1);
    
    D_skiped      = zeros(length(D),1);
    flag_computed = 0;
    
    while l > 0  % loop for eigevalue calculation
        if flag_computed == 1 && norm(D-D_skiped) == 0
            D_skiped = zeros(length(D),1);
            flag_computed = 0;
        end
        [~, t] = max(abs(D-D_skiped));
        CD     = D - D(t); 
        
        z = zeros(l, 1); z(t) = 1; 
        
        % Previous value of g, make inf so at least one iteration occurs
        gp      = inf;
        v2      = v.*v;
        ztv     = z.'*v;
        d       = (CD(t) + r*ztv^2);
        g       = d;
        beta    = 1;
        
        num_iter_step = 0;
        
        while abs((g - gp)/g) > tol.rq      % Modified Rayleigh quotient.     
            % Terminate when relative residual is below tolerance
            % This should be 5n FLOPS, where n = length(z)
            if num_iter_step > 20*pow2(beta) % change step size, when needed, and start from begining
                 z             = zeros(l, 1); z(t) = 1;          
                 beta          = beta +1;
                 v2            = v.*v;
                 ztv           = z.'*v;
                 d             = pow2(1-beta)*(CD(t) + r*ztv^2);
                 g             = d;
                 num_iter_step = 0;
            end
            if beta > 1
                flag_break = 1;
                % disp('Warning: skip shift')
                break;
            end
            num_iter_step = num_iter_step + 1;
            num_iter(l,1) = num_iter(l,1) + 1;
            CDg           = CD - g;
            zv            = v2./CDg;
            ztv           = sum(zv);
            ztz           = sum(zv./CDg);
            d             = pow2(1-beta)*(ztv + r*ztv^2) / ztz;
            gp            = g;
            g             = g + d;
        end
        if flag_break == 1
            D_skiped(t) = D(t);
        end
        if norm(D - D_skiped) ~= 0 && flag_break == 1 
            flag_break = 0;
            continue
        else
            if flag_break == 1 && flag_computed == 0
                disp('Warning: convergence error use eig')
                L(N-l+1:N) = eig(diag(D)+r*(v*v'));
                fprintf('N = %d, l = %d\n',N,length(D));
                break;
            else
                flag_computed = 1;
                
                L(N - l + 1) = D(t) + g; % eigenvalue updating.
                % deflation
                l            = l - 1;
                v            = v ./ sqrt(1 - g./(D - D(t)));
                % Remove the inf/nan from dividing by zero for t
                v(t)         = [];
                D(t)         = [];
                D_skiped(t)  = [];
            end
        end
    end
    
    % eigenvector calculation, if needed
    if nargout == 2
        X = zeros(N, N);
    end
    % Modified Rayleigh quotient update. 
    for l = 1 : N
        CD      = BD - L(l);
        z       = Bv ./ CD;
        Bvtz    = Bv.'*z;
        d       = (Bvtz)*(r*Bvtz + 1)/(z.'*z);
        L(l)    = L(l) + d;
        if nargout == 2
            xl  = Bv ./ (CD - d);

            % Applying Sherman Morrison Woodbury formula and inverse
            % iteration to improve accuracy of the eigenvectors
            if tol.flag_smw_dpr1 
                D       = BD - L(l);
                DinvBv  = Bv ./ D;
                Dinv_xl = xl ./ D;
                rvtBt   = r*Bv.';
                rvtBt_s = rvtBt/(1 + rvtBt*DinvBv);
                x_new   = Dinv_xl - rvtBt_s*Dinv_xl*DinvBv;
                if any(~isfinite(x_new))
                    break
                else
                    xl  = x_new;
                end
            end
            X(:,l)      = xl;
        end
    end
    if nargout == 2
        % Rescale rows of X: matrix entrywise divide by column vector
        X           = X./s;
        % Normalize columns of X
        norms       = sqrt(sum(abs((X.^2)))); % row vector, norms of cols
        X           = X./norms; % Matrix entrywise division with row vector
        XN          = eye(size(LN,1));
        LN(ind)     = L;
        XN(ind,ind) = X;
    else
        LN(ind)     = L;
    end

end
