function data = optViscosities(sys,use_dpr1,tol_sa,a,b,omegas,eta,phis)
% Given a second-order system sys, optimize the viscosities via 
%   Approach 0: only minimize the spectral abscissa
%   Approach 1: minimize the spectral abscissa subject to fixed ellipse
%               constraints
%   Approach 2: maximize the widths of ellipses subject to a stability
%               constraint
%
% INPUT:
%   sys:        struct defining secord-order system, see also makeSysStruct
%   use_dpr1:   use polyeig (false), use dpr1eigmulti (true)
%   tol_sa:     spectral abscissa bound or [] (automatic bound value)
%   a:          Real-valued vector of semi-major axis values (Approach 1)
%   b:          Real-valued vector of semi-minor axis values
%   omegas:     Real-valued vector of frequences to center ellipses at
%   eta:        Nonnegative scalar specifying how far to shift the ellipses
%               The center of the jth ellipse is eta + 1i*omegas(j)
%   phis:       Real-valued vector of weights in [0,1] (Approach 2)
%   
%   APPROACH 0 call:
%   data = optViscosities(sys,use_dpr1,tol_sa,[],[],[],[],[]);
%
%   APPROACH 1 call:
%   data = optViscosities(sys,use_dpr1,tol_sa,a,b,omegas,eta,[])
%
%   APPROACH 2 call:
%   data = optViscosities(sys,use_dpr1,tol_sa,[],b,omegas,eta,phis)
% 
% OUTPUT:
%   data:       struct of data for the optimization
%
% =========================================================================
% |  optViscosities.m                                                     |
% |  Copyright (C) 2020 Tim Mitchell                                      |
% |                                                                       |
% |  This program is free software: you can redistribute it and/or modify |
% |  it under the terms of the GNU Affero General Public License as       |
% |  published by the Free Software Foundation, either version 3 of       |
% |  the License, or (at your option) any later version.                  |
% |                                                                       |
% |  This program is distributed in the hope that it will be useful,      |
% |  but WITHOUT ANY WARRANTY; without even the implied warranty of       |
% |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        |
% |  GNU Affero General Public License for more details.                  |
% |                                                                       |
% |  You should have received a copy of the GNU Affero General Public     |
% |  License along with this program.  If not, see                        |
% |  <http://www.gnu.org/licenses/agpl.html>.                             |
% =========================================================================

%     global ind_m;
%     global flag_warning;
%     flag_warning    = 0;
%     ind_m           = 1;
          
    if isempty(b)
        % Approach 0 - just minimize spectral abscissa 
        a           = [];
        omegas      = [];
        eta         = 0;
        phis        = [];
        fprintf('Running Approach 0...\n');
        approach    = 0;
    elseif isempty(a)
        % Approach 2 - a values can be anything since they are determined
        % by optimization   
        assert(sameSize(omegas,phis),'omegas and phis must be the same size'); 
        fprintf('Running Approach 2...\n');
        approach    = 2;
        a           = ones(size(omegas));
    elseif isempty(phis)
        % Approach 1 - phis can be be anything since they are not used.
        assert(sameSize(a,omegas),'a and omegas must be the same size');
        fprintf('Running Approach 1...\n');
        approach    = 1;
        phis        = ones(size(omegas));
    else
        error('You must provide either a,omegas,eta,[] or [],omegas,eta,phis to determine Approach 1 or 2');
    end
    assert(isscalar(eta) & imag(eta) == 0 & eta >= 0,'eta must be a nonnegative real scalar');
     
    % Offline phase:
    t_start     = tic();
    spec_fns    = setupSpecFns(sys.K,sys.C,sys.M,sys.B,sys.alpha,use_dpr1);
    z_none      = spec_fns.getSpecVZero();
    t_off       = toc(t_start);
    
    % Define semi-minor axes based on undamped spectrum
    if isempty(b)
        b       = 10*max(imag(z_none))/sys.n*ones(size(omegas));  
    end
    n_var       = spec_fns.n_var;
    
    % For these experiments, we'll always start GRANSO from a vector of
    % ones and do at most 50 iterations.
    x0          = ones(n_var,1);
    maxit       = 100;
    mu0         = 10000;

    % Compute spectrum of damped system w/ initial random viscosities
    % If viscosities are nonnegative, spectral abscissa is negative
    z_x0        = spec_fns.getSpec(x0);
    sa_x0       = max(real(z_x0)); 
    % Undamped system must have negative spectral abscissa
    sa_none     = max(real(z_none));
    
    % Don't let the spectral abscissa exceed much beyond that of the
    % randomly damped system (assuming it is stable)
    if isempty(tol_sa)
        tol_sa  = 0.9*min([sa_x0 sa_none]); 
    end

    % Note that if tol_sa is set to aggressively (small), then x0 might be
    % an infeasible point for Approach 2.  If so, the barrier might be
    % infinity, and so optimization cannot from it.  In this case, we would
    % need an initial spectral abscissa minimization phase to ensure we can
    % start a point sufficiently inside the feasible region. For now, let's
    % not set tol_sa too small!
    
    % Setup data and functions for GRANSO 
    opt_fns     = setupOptFns(spec_fns,a,b,omegas,eta,phis,tol_sa);
    if approach == 2
        gransoFn = opt_fns.approach2;
    else
        gransoFn = opt_fns.approach1;
    end
    
    [halt_log_fn,get_log_fn]    = makeHaltLogFunctions(maxit);
    opts.x0                     = x0;
    opts.mu0                    = mu0;
    opts.maxit                  = maxit;
    opts.halt_log_fn            = halt_log_fn;
    opts.quadprog_info_msg      = false;    

    % Online phase:
    t_start = tic(); 
    soln    = granso(n_var,gransoFn,opts);
    t_on    = toc(t_start);
    
    % Compute spectrum of the optimized system
    x_opt   = soln.most_feasible.x;
    z_opt   = spec_fns.getSpec(x_opt);
    
    if approach == 2
        [avals,za]  = opt_fns.getActive2(z_opt);
    elseif approach == 1
        [~,za]      = opt_fns.getActive1(z_opt);
        avals       = a;
    else
        za          = [];
        avals       = [];
    end
    
    % Problem setup data
    data.approach   = approach;
    data.alpha      = sys.alpha;
    data.tol_sa     = tol_sa;
    data.a          = avals;    % These are optimized for approach two
    data.b          = b;
    data.omegas     = omegas;
    data.eta        = eta;
    data.phis       = phis;
    
    % Optimized viscosities and spectra and active eigenvalues
    data.v_opt      = x_opt;
    data.z_none     = z_none;
    data.z_x0       = z_x0;
    data.z_opt      = z_opt;
    data.z_active   = za;
    
    % Metadata
    data.t_off      = t_off;
    data.t_on       = t_on;
    data.soln       = soln;
    data.log        = get_log_fn();
  
end

function tf = sameSize(A,B)
    tf = all(size(A) == size(B));
end
