function [b,b_grad] = barrierFn(f,f_grad,y1,y2,h,c)
% Implementation of barrier function (2.10) and its gradient (2.11)
%
%   INPUT 
%   f           value of f(x) at some point x
%   f_grad      gradient of f(x) at some point x
%   y1,y2       y1 < y2 where the barrier function increases from zero to
%               infinity over the interval from f=y1 to f=y2.
%   h           positive height at which cubic should meet log part
%   c           (0,1) value where cubic and log part should meet
%   OUTPUT
%   b           value of the barrier function
%   b_grad      gradient of the barrier function
%
% =========================================================================
% |  barrierFn.m                                                          |
% |  Copyright (C) 2020 Tim Mitchell                                      |
% |                                                                       |
% |  This program is free software: you can redistribute it and/or modify |
% |  it under the terms of the GNU Affero General Public License as       |
% |  published by the Free Software Foundation, either version 3 of       |
% |  the License, or (at your option) any later version.                  |
% |                                                                       |
% |  This program is distributed in the hope that it will be useful,      |
% |  but WITHOUT ANY WARRANTY; without even the implied warranty of       |
% |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        |
% |  GNU Affero General Public License for more details.                  |
% |                                                                       |
% |  You should have received a copy of the GNU Affero General Public     |
% |  License along with this program.  If not, see                        |
% |  <http://www.gnu.org/licenses/agpl.html>.                             |
% =========================================================================

    % Compute the point y that the barrier transitions from the cubic
    % polynomial to the log-based function
    % Note that y1 and y2 should never change so y is a constant
    if nargin < 5 || isempty(h)
        h   = 1;
    end
    if nargin < 6 || isempty(c)
        c   = 3*h/(3*h + 1);
    end
    y       = y1 + c*(y2 - y1);
    
    % Defaults value of the gradient of the barrier
    b_grad  = zeros(size(f_grad));
    
    if f <= y1
        b       = 0;
    elseif f >= y2
        b       = inf;
    elseif f < y
        t1      = ((1 + 2*h)*y - y1 - 2*h*y2)/((y2 - y)*(y - y1)^3);
        t2      = (y1 + 3*h*y2 - (1 + 3*h)*y)/((y2 - y)*(y - y1)^2);
        b       = t1*(f - y1)^3 + t2*(f - y1)^2;
        b_grad  = 3*t1*(f - y1)^2 + 2*t2*(f - y1);
    else
        b       = h - log((y2 - f)/(y2 - y));
        b_grad  = f_grad/(y2 - f);
    end

end

