function o = ellipseObj(a,b,c)
% This object provides helpful functions regarding an ellipse specified by 
%   a: semi-major axis (real nonnegative scalar)
%   b: semi-minor axis (real nonnegative scalar)
%   c: center (complex scalar)
% See the comments in each nested function.
%
% =========================================================================
% |  ellipseObj.m                                                         |
% |  Copyright (C) 2020 Tim Mitchell                                      |
% |                                                                       |
% |  This program is free software: you can redistribute it and/or modify |
% |  it under the terms of the GNU Affero General Public License as       |
% |  published by the Free Software Foundation, either version 3 of       |
% |  the License, or (at your option) any later version.                  |
% |                                                                       |
% |  This program is distributed in the hope that it will be useful,      |
% |  but WITHOUT ANY WARRANTY; without even the implied warranty of       |
% |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        |
% |  GNU Affero General Public License for more details.                  |
% |                                                                       |
% |  You should have received a copy of the GNU Affero General Public     |
% |  License along with this program.  If not, see                        |
% |  <http://www.gnu.org/licenses/agpl.html>.                             |
% =========================================================================
 
    eta             = real(c);
    omega           = imag(c);
    
    o.distFn        = @distWithGradient;
    o.axisFn        = @axisWithGradient;
    o.distActiveFn  = @distActive;
    o.axisActiveFn  = @axisActive;
   
    function [d,za,idx] = distActive(z)
        % This implements equation (2.4) and the active eigenvalue(s)
        
        % Compute all algebraic distances for each eigenvalue
        dists       = (real(z - c)/a).^2 + (imag(z - c)/b).^2;
        % Get the closest distance (could be ties but unlikely)
        [d,idx]     = min(dists);
        % The active eigenvalue(s), i.e., the closest
        za          = z(idx);
    end

    function [f,g] = distWithGradient(z,X,z_left,Y,zGradFn)
        % This implements equation (2.4) and its gradient 
                   
        [f,za,idx]  = distActive(z);
        % Just in case there are ties (but will basically never happen)
        idx         = idx(1);
        za          = za(1);
        % Get corresponding right and left eigenvectors to compute gradient
        x           = X(:,idx);
        y           = getLeftEigenvector(za,z_left,Y);
        za_grad     = zGradFn(za,x,y);
        % Equation (2.15)
        g           = 2*(   real(za - c)*real(za_grad)/a^2 + ...
                            imag(za - c)*imag(za_grad)/b^2   );
    end

    function [a,za,idx] = axisActive(z)
        % This implements equation (2.8) and the active eigenvalue(s)
        
        % Compute the semi-major axes for each eigenvalue
        re_z        = real(z);
        im_z        = imag(z);
        
        % Compute all the semi-major axes
        a           = b*abs(re_z - eta)./sqrt(b^2 - (im_z - omega).^2);
        % Eigenvalues outside the strip correspond to infinite axis values
        idx         = im_z <= omega - b | im_z >= omega + b;
        a(idx)      = inf;
        % Now get the smallest semi-major axis (could be ties but unlikely)
        [a,idx]     = min(a);
        % The active eigenvalue(s) corresponding to the smallest axis value
        za          = z(idx);
    end

    function [f,g] = axisWithGradient(z,X,z_left,Y,zGradFn)
        % This implements equation (2.8) and its gradient 
        
        [f,za,idx]  = axisActive(z);
        % Just in case there are ties (but will basically never happen)
        idx         = idx(1);
        za          = za(1);
        % Get corresponding right and left eigenvectors to compute gradient
        x           = X(:,idx);
        y           = getLeftEigenvector(za,z_left,Y);
        za_grad     = zGradFn(za,x,y);
        % Equation (2.16) 
        re_za_e     = real(za) - eta;
        im_za_o     = imag(za) - omega;
        b2_term     = b^2 - im_za_o^2;
        term1       = b*sign(re_za_e)*real(za_grad)/sqrt(b2_term);
        term2       = b*abs(re_za_e)*im_za_o*imag(za_grad)*b2_term^-1.5;
        g           = term1 + term2;
    end
      
end

function y = getLeftEigenvector(z,lambda_left,Y)
    [~,indx]    = min(abs(lambda_left - z));
    y           = Y(:,indx);
end