function opt_fns = setupOptFns(spec_fns,a,b,omegas,eta,phis,tol_sa)
% This makes a struct of function handles that GRANSO can use to optimize
% damping viscosities.  Two different but related optimization problems are
% created.
%  
% INPUT:
%   spec_fns:   struct output of setupSpecFns
%
%   a:          Real-valued vector of semi-major axis values (Approach 1)
%   b:          Real-valued vector of semi-minor axis values
%   omegas:     Real-valued vector of frequences to center ellipses at
%   eta:        Nonnegative scalar specifying how far to shift the ellipses
%               The center of the jth ellipse is eta + 1i*omegas(j)
%   phis:       Real-valued vector of weights in [0,1] (Approach 2)
%   tol_sa:     Real-valued scalar for spectral abscissa constraint, i.e,
%               spectral abscissa <= tol_sa should hold for feasibility
%   
%   For Approach 1, just set phis to a vector of 1's of the same size as b.
%   For Approach 2, just set a to a vector of 1's of the same size as b.
%  
% OUTPUT:
%   opts_fns - a struct containing the following fields
%
%   .approach1(v)
%       A function handle to pass to GRANSO for doing Approach 1, where v
%       are the viscosity parameters.  If no ellipses are specified, this
%       just minimizes the spectral abscissa, i.e., Approach 0.
%
%   .approach2(v)
%       A function handle to pass to GRANSO for doing Approach 2, where v
%       are the viscosity parameters
%
%   .getActive1(z)
%       For Approach 1, given eigenvalues z, return the ellipse distances
%       to the closest eigenvalues and return these closest eigenvalues.
%
%   .getActive2(z)
%       For Approach 2, given eigenvalues z, return the semi-major axis
%       values along with the active eigenvalues on the corresponding
%       ellipses 
%
% See also optViscosities
%
% =========================================================================
% |  setupOptFns.m                                                        |
% |  Copyright (C) 2020 Tim Mitchell                                      |
% |                                                                       |
% |  This program is free software: you can redistribute it and/or modify |
% |  it under the terms of the GNU Affero General Public License as       |
% |  published by the Free Software Foundation, either version 3 of       |
% |  the License, or (at your option) any later version.                  |
% |                                                                       |
% |  This program is distributed in the hope that it will be useful,      |
% |  but WITHOUT ANY WARRANTY; without even the implied warranty of       |
% |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        |
% |  GNU Affero General Public License for more details.                  |
% |                                                                       |
% |  You should have received a copy of the GNU Affero General Public     |
% |  License along with this program.  If not, see                        |
% |  <http://www.gnu.org/licenses/agpl.html>.                             |
% =========================================================================
          
    phis        = phis(:);  % ensure phis is a column vector;
    
    getEigPairs = spec_fns.getEigPairs;
    getEigGrad  = spec_fns.getEigGrad;
         
    c           = eta + 1i*omegas;
    n_ell       = length(c);
    ellstruc    = arrayfun(@ellipseObj,a,b,c,'UniformOutput',false);
    
    opt_fns     = struct(                   ...
        'approach1',        @approach1,     ...
        'approach2',        @approach2,     ...
        'getActive1',       @getActive1,    ...
        'getActive2',       @getActive2     );
            
    function [f,f_grad,ci,ci_grad,ce,ce_grad] = approach1(v)
        
        % Precomputations
        [z,X,z_left,Y,C_damped] = getEigPairs(v);
        [sa,sa_grad]            = spectralAbscissa(z,X,z_left,Y,C_damped);
        
        % Objective is minimize the spectral abscissa
        [f,f_grad]          = deal(sa,sa_grad);
        
        % Constraints: bounds, spectral abscissa to tolerance, ellipses
        [b_con,b_grad]      = boundConstraints(v);
        [s_con,s_grad]      = deal(sa - tol_sa,sa_grad);
        if isempty(a) 
            [e_con,e_grad]  = deal(zeros(0,1)); 
        else
            [e_con,e_grad]  = ellipseConstraints(z,X,z_left,Y,C_damped);
        end
        ci                  = [b_con; s_con; e_con];
        ci_grad             = [b_grad s_grad e_grad];
        
        % no equality constraints for this problem
        ce                  = [];
        ce_grad             = [];
    end

    function [f,f_grad,ci,ci_grad,ce,ce_grad] = approach2(v)
        
        % Precomputations
        [z,X,z_left,Y,C_damped] = getEigPairs(v);
        [sa,sa_grad]            = spectralAbscissa(z,X,z_left,Y,C_damped);
        [e,e_grad]              = ellipseObjective(z,X,z_left,Y,C_damped);
        [beta,beta_grad]        = barrierFn(sa,sa_grad,tol_sa,eta);
    
        % Objective is max(e - beta) so flip signs for minimization
        f                   = beta - e;
        f_grad              = beta_grad - e_grad;
        
        % Constraints: bounds, spectral abscissa to tolerance
        [b_con,b_grad]      = boundConstraints(v);
        [s_con,s_grad]      = deal(sa - tol_sa,sa_grad);
        ci                  = [b_con; s_con];
        ci_grad             = [b_grad s_grad];
        
        % no equality constraints for this problem
        ce                  = [];
        ce_grad             = [];
    end

    function [f,g] = spectralAbscissa(z,X,z_left,Y,C_damped)
        % Objective: minimize the spectral abscissa
        [~,indx]            = max(real(z));
        z_abs               = z(indx);
        x                   = X(:,indx);
        y                   = getLeftEigenvector(z_abs,z_left,Y);
        
        % the above seems okay (at leaast for nonsymmetric)
        z_abs_grad          = getEigGrad(z_abs,x,y,C_damped);
        f                   = real(z_abs);
        g                   = real(z_abs_grad); 
    end

    function [f,g] = boundConstraints(v)
        % damping viscosities must be nonnegative
        f                   = -v;
        g                   = -eye(length(v));
    end

    function [f,g] = ellipseConstraints(z,X,z_left,Y,C_damped)
        % For Approach 1
        
        zGradFn     = @(z,x,y) getEigGrad(z,x,y,C_damped);
        
        distFn      = @(j) ellstruc{j}.distFn(z,X,z_left,Y,zGradFn);
        [d,d_grad]  = arrayfun(distFn,1:n_ell,'UniformOutput',false);
        % Convert cell arrays to a column vector of distances and a matrix
        % of the corresponding gradients
        d           = cell2mat(d).';
        d_grad      = cell2mat(d_grad);
        
        % Now transform the distances into <= 0 inequality constraints
        f           = 1 - d;
        g           = -d_grad;
        
        % We could also consider using a max here to fold down the multiple
        % ellipse constraints into a single constraint
%         [f,idx]     = max(f);
%         g           = g(:,idx);
    end

    function [f,g] = ellipseObjective(z,X,z_left,Y,C_damped)
        % For Approach 2
        
        zGradFn     = @(z,x,y) getEigGrad(z,x,y,C_damped);
        
        axisFn      = @(j) ellstruc{j}.axisFn(z,X,z_left,Y,zGradFn);
        [f,f_grad]  = arrayfun(axisFn,1:n_ell,'UniformOutput',false);
        % Convert cell arrays to a column vector of distances and a matrix
        % of the corresponding gradients
        f           = cell2mat(f).';
        f_grad      = cell2mat(f_grad);
        
        % Objective and gradient are weighted sums via phis
        f           = sum(f.*phis);
        g           = f_grad*phis;
    end

    function [d,za] = getActive1(z)
        activeFn    = @(j) ellstruc{j}.distActiveFn(z);
        [d,za]      = arrayfun(activeFn,1:n_ell,'UniformOutput',false);
        d           = cell2mat(d).';
    end

    function [a,za] = getActive2(z)
        activeFn    = @(j) ellstruc{j}.axisActiveFn(z);
        [a,za]      = arrayfun(activeFn,1:n_ell,'UniformOutput',false);
        a           = cell2mat(a).';
    end

end

function y = getLeftEigenvector(z,lambda_left,Y)
    [~,indx]    = min(abs(lambda_left - z));
    y           = Y(:,indx);
end