function o = setupSpecFns(K,C,M,B,alpha,use_dpr1)
% Set up functions for computing eigenvalues, eigenvectors, and eigenvalue
% gradients using either polyeig or the dpr1 code.
%
% =========================================================================
% |  setupSpecFns.m                                                       |
% |  Copyright (C) 2020 Tim Mitchell                                      |
% |                                                                       |
% |  This program is free software: you can redistribute it and/or modify |
% |  it under the terms of the GNU Affero General Public License as       |
% |  published by the Free Software Foundation, either version 3 of       |
% |  the License, or (at your option) any later version.                  |
% |                                                                       |
% |  This program is distributed in the hope that it will be useful,      |
% |  but WITHOUT ANY WARRANTY; without even the implied warranty of       |
% |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        |
% |  GNU Affero General Public License for more details.                  |
% |                                                                       |
% |  You should have received a copy of the GNU Affero General Public     |
% |  License along with this program.  If not, see                        |
% |  <http://www.gnu.org/licenses/agpl.html>.                             |
% =========================================================================

    if use_dpr1
        dpr1_tol    = getDpr1Tols();
        dpr1_data   = transformEigensystem(K,M,B,alpha);
        dpr1Fn      = @(r) dpr1eigmulti(dpr1_data,r,dpr1_tol);
        z_none      = dpr1_data.D;
    else
        z_none      = [];
    end
   
%     one_call        = isreal(K) && isreal(C) && isreal(M);
%     if ~one_call
%         one_call    = ishermitian(K) && ishermitian(C) && ishermitian(M); 
%     end
    
    n_var           = size(B,2);
    o.n             = length(K);
    o.n_var         = n_var;
    o.getSpec       = @getSpectrum;
    o.getSpecVZero  = @getSpectrumUndamped;
    o.getEigPairs   = @getEigenpairs;
    o.getEigGrad    = @eigenvalueGradient;
       
    function spec = getSpectrum(v)
        if use_dpr1   
            spec    = dpr1Fn(v);
        else
            spec    = polyeig(K,getDampedC(v),M);
        end
    end

    function spec = getSpectrumUndamped(varargin)
        if isempty(z_none)
            z_none  = getSpectrum(zeros(n_var,1));
        end
        spec        = z_none;
    end
    
    function [z,X,z_left,Y,C_damped] = getEigenpairs(v)
        % Precomputations for the objective and the constraint
        C_damped    = getDampedC(v);
        if use_dpr1 
            [X,z]   = dpr1Fn(v);
        else
            [X,z]   = polyeig(K,C_damped,M);
        end
        Y           = X;
        z_left      = conj(z);
    end

    function grad = eigenvalueGradient(z,x,y,C_damped)
        % compute the eigenvalue gradient of the damped system defined by
        % the matrices K,C+damping,M, where K and M are fixed.
        scalar      = -z/(y'*(C_damped + 2*z*M)*x);
        Bhy         = B'*y;
        Bhx         = B'*x;
        grad        = conj(Bhy)*Bhx.';  
        grad        = diag(grad);
        grad        = scalar*grad;
    end

    function C_damped = getDampedC(v)
        C_damped    = C + B*diag(v)*B';
    end
    
end