function data = runEigScaling(n_vals,conf_idx,vs)
%
% =========================================================================
% |  runEigScaling.m                                                      |
% |                                                                       |
% |  This program is free software: you can redistribute it and/or modify |
% |  it under the terms of the GNU Affero General Public License as       |
% |  published by the Free Software Foundation, either version 3 of       |
% |  the License, or (at your option) any later version.                  |
% |                                                                       |
% |  This program is distributed in the hope that it will be useful,      |
% |  but WITHOUT ANY WARRANTY; without even the implied warranty of       |
% |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        |
% |  GNU Affero General Public License for more details.                  |
% |                                                                       |
% |  You should have received a copy of the GNU Affero General Public     |
% |  License along with this program.  If not, see                        |
% |  <http://www.gnu.org/licenses/agpl.html>.                             |
% =========================================================================

    k       = length(n_vals);
    
    if nargin < 3
        % If viscosities are provided, generate random ones
        vs  = 0.1 + rand(3,k);
    end
    if size(vs,2) == 1
        vs  = repmat(vs,1,k);
    end
    assert(norm(size(vs) - [3, k]) == 0,'vs must be 3 by 1 or 3 by length(n_vals)');

    t_dpr1_off  = zeros(k,1);
    t_dpr1_on   = zeros(k,1);
    z_dpr1      = cell(k,1);  
    r_dpr1      = cell(k,1);
    t_eig_off   = zeros(k,1);
    t_eig_on    = zeros(k,1);
    z_eig       = cell(k,1);
    t_poly      = zeros(k,1);
    z_poly      = cell(k,1);
    t_quad      = zeros(k,1);
    z_quad      = cell(k,1);
    
    for j = 1:k
        n = n_vals(j);
        v = vs(:,j);
        
        fprintf('Running n = %5d ... ',n); 
        t_start = tic;
        
        sys     = getExample4(n,conf_idx);
        K       = sys.K;
        C       = sys.C;
        M       = sys.M;
        B       = sys.B;
   
        [z_dpr1{j},r_dpr1{j},t_dpr1_on(j),t_dpr1_off(j)]  = runDpr1(v,sys);
        [z_eig{j},t_eig_on(j),t_eig_off(j)]     = runEigA(v,K,C,M,B);
        [z_poly{j},t_poly(j)]                   = runPoly(v,K,C,M,B);
        [z_quad{j},t_quad(j)]                   = runQuad(v,K,C,M,B);
        
        fprintf('finished in %8.1f seconds.\n',toc(t_start));
    end
    
    times.dpr1_off  = t_dpr1_off;
    times.dpr1_on   = t_dpr1_on;
    times.eig_off   = t_eig_off;
    times.eig_on    = t_eig_on;
    times.polyeig   = t_poly;
    times.quadeig   = t_quad;
    
    evalues.dpr1    = z_dpr1;
    evalues.eig     = z_eig;
    evalues.polyeig = z_poly;
    evalues.quadeig = z_quad;
    
    data.n_vals     = n_vals;
    data.conf_idx   = conf_idx;
    data.vs         = vs;
    data.times      = times;
    data.evalues    = evalues;
    data.res_dpr1   = r_dpr1;

end

function [z,r,t_on,t_off] = runDpr1(v,sys)
    t_start     = tic;
    dpr1_data   = transformEigensystem(sys.K,sys.M,sys.B,sys.alpha);
    dpr1Fn      = @(r) dpr1eigmulti(dpr1_data,r,getDpr1Tols());
    t_off       = toc(t_start);
    
    t_start     = tic;
    [X,z]       = dpr1Fn(v);
    t_on        = toc(t_start);
    r           = computeResiduals(sys,v,z,X);
end

function [z,t_on,t_off] = runEigA(v,K,C,M,B)
    t_start     = tic;
    [Phi,O2]    = eig(K,M,'vector');
    Omega       = diag(sqrt(O2));
    t_off       = toc(t_start);
    
    % We could move some of the matrix construction to offline but I think
    % it is marginal compared to calling eig
    t_start     = tic;
    C_damped    = C + B*diag(v)*B';
    A           = [ zeros(size(Omega)) Omega; -Omega (-Phi.'*C_damped*Phi)];
    [X,z]       = eig(A,'vector');
    t_on        = toc(t_start);
end

function [z,t] = runPoly(v,K,C,M,B)
    t_start     = tic;
    [X,z]       = polyeig(K,C + B*diag(v)*B',M);
    t           = toc(t_start);
end

function [z,t] = runQuad(v,K,C,M,B)
    t_start     = tic;
    [X,z]       = quadeig(M,C + B*diag(v)*B',K);
    t           = toc(t_start);
end

function [res] = computeResiduals(sys,v,L,X)
    res = zeros(size(L,1),1);
    for ii = 1 : size(L,1)
        Q          = L(ii)^2*sys.M+L(ii)*(sys.C+sys.B*diag(v)*sys.B.')+sys.K;
        res(ii)    = norm(Q*X(:,ii));
    end
end
